#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Smartphone Radar
# Developed by Travis Cook 
# www.frightideas.com

import os
import sys
import socket
import indigo
import time
import re
import subprocess
import signal

kSocketPort = 1514
kSocketBufferSize = 1024
kSocketTimeout = 1
kSNMPKillTime = 10
kSnmpInterval = 60
kMinimumIndigoVersion = '5.0.3'		#	FIX FIX FIX CHANGE THIS TO 5.0.3 BEFORE SHIP


################################################################################
class iPresence(object):

	def enum(self,**enums):
		return type('Enum', (), enums)
		
	########################################
	def __init__(self, plugin):

		self.States = self.enum(STARTUP=1,HOLD=2,INIT=3,  SNMP_INIT=10,SNMP_IDLE=11,SNMP_RUN=12,SNMP_POLL=13,  SYSLOG_INIT=20,SYSLOG_POLL=21)
		self.state = self.States.STARTUP
		
		self.sock = None
		self.plugin = plugin
		self.shutdown = False
		self.configRead = False
		self.showRouterLog = False
		self.socketOpen = False
		self.devList = {}
		self.triggerList = {}
		self.createVariables = False
		self.numberOfPhones = 0
		self.AllPhonesHome = True
		self.AllPhonesAway = False
		self.AllPhonesOnVacation = False
		self.varAllPhonesHome = None
		self.varAllPhonesAway = None
		self.varAllPhonesOnVacation = None
		self.processes = []
		

	def __del__(self):
		pass
			
	def startup(self):
		self.plugin.debugLog(u"iPresence startup called")
			
			
	########################################
	# Socket Subs
	
	def openSocket(self):
		self.plugin.debugLog(u"openSocket called")
		try:
			self.plugin.debugLog(u"Opening port %s" % kSocketPort)
			self.sock=socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
			#self.sock=socket.socket(socket.AF_UNIX, socket.SOCK_RAW)
			self.sock.bind(("0.0.0.0",kSocketPort))
			self.sock.settimeout(kSocketTimeout)
			# Get socket IP Address
			self.socketOpen = True

		except socket.error:
			self.plugin.debugLog(u"Socket Setup Error: (%s) %s" % ( sys.exc_info()[1][0], sys.exc_info()[1][1] ))

	def closeSocket(self):
		if self.sock != None:
			self.plugin.debugLog(u"closeSocket called")
			self.sock.close
			self.sock = None
			
	def getSocketIP(self):
		if self.socketOpen == True:
			return socket.gethostbyname(socket.gethostname())
		else:
			return u"Click Ok and ReOpen to get this"
		
	def getSocketPort(self):
		return kSocketPort		
		
		
	########################################
	# Indigo Version Check
	#
	# Returns True if Indigo's version is equal to or higher than
	# the string supplied.  minStr format example: '5.0.2'
	#		
# 	def indigoVersionIsAtLeast(self, minStr):
# 				
# 		(v1,v2,v3) = indigo.server.version.split('.')
# 		(m1,m2,m3) = minStr.split('.')
# 		
# 		verNow = (int(v1)*1000) + (int(v2)*100) + int(v3)
# 		verMin = (int(m1)*1000) + (int(m2)*100) + int(m3)
# 		
# 		if verNow >= verMin:
# 			return True
# 		else:
# 			return False
			
			
	########################################
	# Configuration
	#
	# Reads the indigo plugin Dict into our own variables.
	# Returns True on success
	#
	def getConfiguration(self,valuesDict):
		#Set debug on / off
		
		self.plugin.debugLog(u"getConfiguration start")

		try:
			# Get setting of Debug checkbox
			if valuesDict[u'showDebugInfo'] == True:
				self.plugin.debug = True
			else:
				self.plugin.debug = False
				
			# Get setting of Show Router Log checkbox
			if valuesDict[u'showRouterLog'] == True:
				self.showRouterLog = True
			else:
				self.showRouterLog = False
				
			# Get setting of Create Variables checkbox
			if valuesDict[u'createVariables'] == True:
				self.createVariables = True
			else:
				self.createVariables = False
				
			self.minutesTillAway = int(valuesDict[u'minutesTillAway'])
			if self.minutesTillAway < 2:
				self.minutesTillAway = 2
				
			self.minutesTillVacation = int(valuesDict[u'hoursTillVacation']) * 60	
				
			return True

		except:

			self.plugin.debugLog(u"Error reading plugin configuration. (happens on very first launch)")
		
		return False			
	
		
	########################################
	# Variable Subs
	
	# Updates indigo variable instance var with new value varValue
	#
	def updateVariable(self, var, varValue):
		#self.plugin.debugLog(u"Variable: %s" % var)
		if self.createVariables == False:
			return
		var.value = varValue
		var.replaceOnServer()
	
	# First searches for a variable with name varName.  If it doesn't 
	# exist it creates it in folder instance varFolder.  The variable 
	# value is then set to intialValue and the variable instance is returned.
	#
	def getOrCreateVariable(self, varName, initialValue, varFolder):
		if initialValue == True:
			initialValue = "true"
		elif initialValue == False:
			initialValue = "false"
			
		for var in indigo.variables:
			if var.name == varName:
				var.value = initialValue
				var.replaceOnServer()
				return var
		return indigo.variable.create(varName, value=initialValue, folder=varFolder)
	
	
	# Looks for our home variable folder, if it doesn't exist it's created.
	# Our global variables and phone variables are then created and initialized.
	#
	def createVariabless(self):
		if self.createVariables == False:
			return
		self.plugin.debugLog(u'Entered createVariables')
		ourVariableFolder = None
		for folder in indigo.variables.folders:
			if folder.name == u"Smartphone Radar":
				ourVariableFolder = folder
		if ourVariableFolder == None:
			ourVariableFolder = indigo.variables.folder.create(u"Smartphone Radar")
		
		#create our global indigo variables
		self.varAllPhonesHome = self.getOrCreateVariable(u"AllPhonesHome",self.AllPhonesHome,ourVariableFolder)
		self.varAllPhonesAway = self.getOrCreateVariable(u"AllPhonesAway",self.AllPhonesAway,ourVariableFolder)
		self.varAllPhonesOnVacation = self.getOrCreateVariable(u"AllPhonesOnVacation",self.AllPhonesOnVacation,ourVariableFolder)
		
		# Create a "Home" variable for each phone.  
		for phone in self.devList:
			# First remove any invalid characters from the device name
			cleanName = re.sub(r'[^a-zA-Z0-9_-]+', '', self.devList[phone]['name'])
			# Create or find the instance of the variable with the clean name
			self.devList[phone]['varHomeRef'] = self.getOrCreateVariable(''.join([cleanName,'Home']),True,ourVariableFolder)
		
		
		
	def getSnmpSearchStrings(self, macAdr):

		srch = []
		
		if macAdr.count(':') != 5:
			return []
		
		macAdr = macAdr.lower()
		
		dec = ''
		try:
			for byte in macAdr.split(':'):
				dec += str(int(byte, 16)) + '.'
		except:
			indigo.server.log("Error parsing MAC address " + macAdr, isError=True)
	
		srch.append(dec.rstrip('.'))
		srch.append(macAdr)
		srch.append(macAdr.replace(':',' '))
		
		self.plugin.debugLog(u"Search List: %s" % srch)
		
		return srch
		
	########################################
	# Device Start / Stop Subs
		
	# Called by Indigo when a phone device is added 
	# (called at every plugin startup for each device)
	#
	def deviceStart(self, dev):
		self.plugin.debugLog(u"deviceStart called.  Adding device %s." % dev.name)
		self.plugin.debugLog(u"Device: %s" % dev)
		#self.phoneList[dev.id] = {"phoneID":dev.pluginProps['phoneID']}
		self.devList[dev.id] = {'ref':dev,'name':dev.name,'phoneID':dev.pluginProps['phoneID'].lower(),'lastMinuteInc':time.time(), 'varHomeRef':None}
		#self.plugin.debugLog(u"DevList: %s" % self.devList)
		indigo.server.log(u"Watching for smartphone '%s' at address '%s'." % (self.devList[dev.id]['name'],self.devList[dev.id]['phoneID']))
		self.devList[dev.id]['snmpSearchStrings'] = self.getSnmpSearchStrings(self.devList[dev.id]['phoneID'])
		self.devList[dev.id]['ref'].updateStateOnServer(key="state", value='home')
		self.devList[dev.id]['ref'].updateStateOnServer(key="lastSeen", value=1)
		self.numberOfPhones += 1
		# Tell concurrent loop to reread our config
		self.configRead = False
	
	# Called by Indigo when a phone device is deleted
	# (called at every plugin stop for each device)
	#
	def deviceStop(self, dev):
		if dev.id not in self.devList:
			return
		self.plugin.debugLog(u"deviceStop called.  Removing device %s." % dev.name)
		del self.devList[dev.id]
		self.plugin.debugLog(u"DevList is now: %s" % self.devList)
		self.numberOfPhones -= 1


	########################################
	# Trigger Start / Stop Subs
		
	# Called by Indigo when a trigger is added 
	# (called at every plugin startup for each Events.xml trigger instance in Indigo)
	#
	def triggerStart(self, trigger):
		self.plugin.debugLog(u"triggerStart called.  Adding trigger %s." % trigger.name)
		#self.plugin.debugLog(u"Trigger: %s" % trigger)
		self.triggerList[trigger.id] = trigger
	
	# Called by Indigo when a trigger is removed or disabled
	# (called at every plugin stop for each trigger instance)
	#
	def triggerStop(self, trigger):
		self.plugin.debugLog(u"triggerStop called.  Removing trigger %s." % trigger.name)
		del self.triggerList[trigger.id]

	# Executes all triggers that match the supplied eventID
	#
	def triggerExecuteEvent(self, eventId):
		for trig in self.triggerList:
			if self.triggerList[trig].pluginTypeId == eventId:
				indigo.trigger.execute(self.triggerList[trig])


	########################################
	# Phone Device Timer and Status Subs
	#
	def setGlobalPhoneVariables(self):
	
		# See if all phones are on vacation true or false
		phonesOnVacation = self.countPhonesWithState(u'onVacation')
		if self.AllPhonesOnVacation == False and (phonesOnVacation == self.numberOfPhones):
			self.AllPhonesOnVacation = True
			indigo.server.log(u"All smartphones are now on Vacation.")
			self.triggerExecuteEvent(u'allPhonesOnVacation')
			self.updateVariable(self.varAllPhonesOnVacation,u'true')
		elif self.AllPhonesOnVacation == True and (phonesOnVacation != self.numberOfPhones):
			self.AllPhonesOnVacation = False
			self.updateVariable(self.varAllPhonesOnVacation,"false")	
		
		# If all phones are on vacation then skip the rest of the tests
		if self.numberOfPhones == phonesOnVacation:
			return
			
		# See if all phones are away true or false
		phonesAway = self.countPhonesWithState('away')
		if self.AllPhonesAway == False and (phonesAway == self.numberOfPhones - phonesOnVacation):
			self.AllPhonesAway = True
			indigo.server.log(u"All smartphones are now Away.")
			self.triggerExecuteEvent(u'allPhonesAway')
			self.updateVariable(self.varAllPhonesAway,"true")	
		elif self.AllPhonesAway == True and (phonesAway != self.numberOfPhones - phonesOnVacation):
			self.AllPhonesAway = False
			self.updateVariable(self.varAllPhonesAway,"false")	
		
		# See if all phones are home true or false
		phonesAtHome = self.countPhonesWithState('home')
		if self.AllPhonesHome == False and (phonesAtHome == self.numberOfPhones - phonesOnVacation):
			self.AllPhonesHome = True
			indigo.server.log(u"All smartphones are now Home.")
			self.triggerExecuteEvent(u'allPhonesHome')
			self.updateVariable(self.varAllPhonesHome, "true")	
		elif self.AllPhonesHome == True and (phonesAtHome != self.numberOfPhones - phonesOnVacation):
			self.AllPhonesHome = False
			self.updateVariable(self.varAllPhonesHome, "false")		
	
	# Goes through our phone list and returns the number of phones with
	# states that match the supplied state argument
	#
	def countPhonesWithState(self,phoneState):
		num = 0
		for phone in self.devList:
			if self.devList[phone]['ref'].states['state'] == phoneState:
				num += 1
		return num
	
	# Called when a phone's activity is seen on the network
	# Requires the phone's device ID as an argument
	#
	def foundPhone(self,phone):
		self.plugin.debugLog(u'"%s" network activity detected.' % self.devList[phone]['name'])	
		self.devList[phone]['lastMinuteInc'] = time.time()
		self.devList[phone]['ref'].updateStateOnServer(key="lastSeen", value=0)
		if self.devList[phone]['ref'].states['state'] != 'home':
			# EVENT A Phone has come home
			self.devList[phone]['ref'].updateStateOnServer(key="state", value='home')
			self.updateVariable(self.devList[phone]['varHomeRef'], "true")	
			indigo.server.log(u'"%s" state is now Home.' % self.devList[phone]['name'])
			self.setGlobalPhoneVariables()

	# Called when a phone hasn't been seen on the network in the last minute
	# Requires the phone's device ID as an argument
	#
	def incPhoneLastSeenTimer(self,phone):
		#indigo.trigger.execute(u'allPhonesHome', ignoreConditions=False)
		self.plugin.debugLog(u"Incrementing lastSeen timer of '%s'" % self.devList[phone]['name'])	
		lastSeen = self.devList[phone]['ref'].states['lastSeen']
		lastSeen += 1
		self.devList[phone]['ref'].updateStateOnServer(key="lastSeen", value=lastSeen)
		
		if lastSeen == self.minutesTillAway:
			self.devList[phone]['ref'].updateStateOnServer(key="state", value='away')
			self.updateVariable(self.devList[phone]['varHomeRef'], "false")	
			
			# EVENT A Phone is now Away
			indigo.server.log(u'"%s" state is now Away.' % self.devList[phone]['name'])
			self.setGlobalPhoneVariables()
					
		if (self.minutesTillVacation > 0) and (lastSeen == self.minutesTillVacation):
			# EVENT A Phone is now on Vacation
			self.devList[phone]['ref'].updateStateOnServer(key="state", value='onVacation')
			indigo.server.log(u'"%s" state is now On Vacation.' % self.devList[phone]['name'])
			self.setGlobalPhoneVariables()

	# Loops through our list of phone devices to see if their minute counter
	# needs to be updated.  Called from the concurrent thread.
	#
	def checkPhoneTimers(self):
		# Check time, see if a minute has passed
		timeNow = time.time()
		# Loop through each phone and compare their last inc time to the current time
		for phone in self.devList:
			if timeNow >= self.devList[phone]['lastMinuteInc'] + 60:
				# A minute passed for this phone.  Store current time and call IncPhone.
				self.devList[phone]['lastMinuteInc'] = timeNow
				self.incPhoneLastSeenTimer(phone)
	
	
	# Loops through our snmp process list and kills everything
	#		
	def killProcesses(self):
		for p in self.processes:
			p[0].stdout.close()
			os.kill(p[0].pid, signal.SIGKILL)
		self.processes = []
			
		
		
	########################################
	# Concurrent Thread Start / Stop
	#
	def startComm(self):
		self.plugin.debugLog(u"startComm called")
		
		
		# While Indigo hasn't told us to shutdown
		while self.shutdown == False:
		
			timeNow = time.time()
				
			#####################
			# STARTUP state
			# Here we check if our version of Indigo is high enough
			if self.state == self.States.STARTUP:
				self.plugin.debugLog(u"STATE: Startup")
				
				self.state = self.States.INIT
				
# 				if self.indigoVersionIsAtLeast(kMinimumIndigoVersion):
# 					self.state = self.States.INIT
# 				else:
# 					indigo.server.log("This plugin requires Indigo version %s or later." % kMinimumIndigoVersion, isError=True)
# 					self.state = self.States.HOLD

			#####################			
			# HOLD state
			# Indigo version isn't high enough, do nothing
			elif self.state == self.States.HOLD:
				self.plugin.sleep(1)
				
			#####################			
			# INIT state
			# Make sure plugin prefs are good, check which method we're using for detection	
			elif self.state == self.States.INIT:
				self.plugin.debugLog(u"STATE: Init")
			
				# Read indigo pluginPrefs into our own variables
				# Keep trying to read until it works.  On the very first startup of the plugin the
				# pluginPrefs Dict is not valid right away, we must poll till it is.
				if self.configRead == False:
					if self.getConfiguration(self.plugin.pluginPrefs) == True:
						self.configRead = True
						if self.createVariables == True:
							self.createVariabless()
							
						# Close the socket incase we already opened it and we're 
						# just here because the prefs were changed.
						self.closeSocket()
							
						if self.plugin.pluginPrefs[u'configSyslogOrSNMP'] == u'syslog':
							self.state = self.States.SYSLOG_INIT
						else:
							self.state = self.States.SNMP_INIT

			#####################			
			# SNMP_INIT state
			# 
			elif self.state == self.States.SNMP_INIT:
				self.plugin.debugLog(u"STATE: SNMP Init")
				self.snmpNextRun = timeNow + 5
				self.state = self.States.SNMP_IDLE
			
			#####################			
			# SNMP_IDLE state
			# 
			elif self.state == self.States.SNMP_IDLE:
				#self.plugin.debugLog(u"STATE: SNMP Idle")
				if timeNow > self.snmpNextRun:
					self.state = self.States.SNMP_RUN
					self.snmpNextRun = timeNow + kSnmpInterval
				
				if self.configRead == False:
					self.state = self.States.INIT
					
			#####################			
			# SNMP_RUN state
			# 
			elif self.state == self.States.SNMP_RUN:
				self.plugin.debugLog(u"STATE: SNMP Run")
				self.state = self.States.SNMP_POLL
				self.snmpKillTimer = timeNow + kSNMPKillTime
				#p = subprocess.Popen(['digitemp', '-c', kDigiTempConf, '-q', '-i', '-s', self.scanPort], stdout=subprocess.PIPE)
				
				routers = self.plugin.pluginPrefs[u'snmpAddress'].split(',')
				communities = self.plugin.pluginPrefs[u'snmpCommunity'].split(',')
				
				for i in range(len(routers)):
					snmpAdr = routers[i].strip()
					snmpComm = communities[i].strip()
					self.plugin.debugLog(u"Polling Airport: *%s*, Pwd: *%s*" % (snmpAdr, snmpComm) )
					#self.plugin.debugLog(u"Polling Airport: %s" % snmpAdr )
					cmd = 'snmpwalk', '-m', 'AIRPORT-BASESTATION-3-MIB', '-Os', '-v', '2c', '-c', '%s' % snmpComm, '%s' % snmpAdr, '1.3.6.1.4.1.63.501.3.2'
					self.plugin.debugLog("SNMP Command: %s" % ' '.join(map(str,cmd)) )
					p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
					self.processes.append((p,snmpAdr))
				
				
			#####################			
			# SNMP_POLL state
			# 
			elif self.state == self.States.SNMP_POLL:
				self.plugin.debugLog(u"STATE: SNMP Poll")
				
				if len(self.processes) == 0:
					self.state = self.States.SNMP_IDLE
				else:
				
					# Check if the processes have exceeded our time limit
					# Kill it if it has
					if timeNow > self.snmpKillTimer:
						self.killProcesses()
						indigo.server.log("Some SNMP processes timed out.  The address is wrong or the router is unreachable.", isError=True)
						self.state = self.States.SNMP_IDLE
					else:
						for p in self.processes:
							p[0].poll()
							if p[0].returncode != None:
								output = p[0].stdout.read()
								p[0].stdout.close()
								
								output = output.lower()
								
								if 'timeout' in output:
									# Gets this if the community string does not exist
									indigo.server.log("The SNMP request from %s timeout out.  Wrong password?" % p[1], isError=True)
								elif 'no such object' in output:
									# Gets this if the community string is there, but doesn't contain wireless info
									indigo.server.log("The SNMP output from %s does not contain any wireless info.  Wrong password?" % p[1], isError=True)
								elif 'wirelessnumber' not in output:
									indigo.server.log("The SNMP output from %s does not contain any wireless info.  Wrong password?" % p[1], isError=True)

								for phone in self.devList:
									for search in self.devList[phone]['snmpSearchStrings']:
										self.plugin.debugLog(u"Searching SNMP data for *%s*" % search)
										if search in output:
											self.foundPhone(phone)
											break
											
								if self.showRouterLog == True:
									indigo.server.log(u"SNMP Output for %s:\n" % p[1] + output)

								self.processes.remove(p)
					

			#####################			
			# SYSLOG_INIT state
			# Open the syslog socket to accept the log from the router
			elif self.state == self.States.SYSLOG_INIT:
				self.plugin.debugLog(u"STATE: syslog Init")
				# Open the syslog socket
				self.openSocket()
				self.state = self.States.SYSLOG_POLL

			#####################			
			# SYSLOG_POLL state
			# Check the socket for data, check the timers
			elif self.state == self.States.SYSLOG_POLL:
				#self.plugin.debugLog(u"STATE: syslog Poll")
				
				if self.configRead == False:
					self.state = self.States.INIT
					
				try:
					# sock.recvfrom seems to only return complete log strings.  One log entry per return.
					# Which is good, because log entries are not delimitted with any \n or \r characters.
					# If responses were coming in with partial log entries and buffering was required
					# to concatenate full strings it would be tricky to determine where one entry begins
					# and the other ends.
					data,addr=self.sock.recvfrom(kSocketBufferSize)
					
					# Show the log if the user wants
					if self.showRouterLog == True:
						indigo.server.log(u"Router Log: %s" % data)
					
					#convert log to lowercase for easier matching to any MAC addresses we stored in lower case
					data = data.lower()
					
					# Scan the data we received for matches to our phones IP Addresses
					for phone in self.devList:
						if self.devList[phone]['phoneID'] in data:
							self.foundPhone(phone)
	
				except socket.timeout:
					# No data was received, socket timed out or bailed for some other reason
					# self.plugin.debugLog(u"Socket Timeout")
					pass
				except:
					self.plugin.debugLog(u"Socket recvfrom exception")
					pass
						
						
			# These run in all states
			self.checkPhoneTimers()
			# Check the phone timers	
			self.plugin.sleep(.5)
		
		# self.shutdown is now true
		self.plugin.debugLog(u"startComm Exit")
		self.closeSocket()
		self.killProcesses()

	def stopComm(self):
		self.plugin.debugLog(u"stopConcurrentThread called")
		self.shutdown = True
