#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Smartphone Radar
# Developed by Travis Cook 
# www.frightideas.com

import os
import sys

from iPresence import iPresence

# Note the "indigo" module is automatically imported and made available inside
# our global name space by the host process.

################################################################################
class Plugin(indigo.PluginBase):
	########################################
	def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs):
		indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)

		self.iPresence = iPresence(self)	
		self.debug = False

	def __del__(self):
		indigo.PluginBase.__del__(self)

	########################################
	def startup(self):
		
		try:		
			if valuesDict[u'showDebugInfo'] == True:
				self.debug = True
			else:
				self.debug = False
		except:
			pass
		
		self.debugLog(u"startup called")
		
		self.iPresence.startup()
		
		# Open the syslog socket
		# self.iPresence.openSocket()
		
		# Get the IP Address and Port of the socket for viewing in the plugin config
		#self.pluginPrefs[u"ourIPAddress"] = self.iPresence.getSocketIP()
		#self.pluginPrefs[u"ourPort"] = self.iPresence.getSocketPort()
		
	def shutdown(self):
		self.debugLog(u"shutdown called")
		# self.iPresence.closeSocket()


	######################
	def deviceStartComm(self, dev):
		self.debugLog(u"<<-- entering deviceStartComm: %s (%d - %s)" % (dev.name, dev.id, dev.deviceTypeId))
		self.iPresence.deviceStart(dev)
		self.debugLog(u"exiting deviceStartComm -->>")

	def deviceStopComm(self, dev):
		self.debugLog(u"<<-- entering deviceStopComm: %s (%d - %s)" % (dev.name, dev.id, dev.deviceTypeId))
		self.iPresence.deviceStop(dev)
		self.debugLog(u"exiting deviceStopComm -->>")
		
	#def deviceUpdated(self, origDev, newDev):
	#	indigo.server.log(u"<<-- entering deviceUpdated: %s" % origDev.name)	
	#	self.iPresence.deviceStop(origDev)
	#	self.iPresence.deviceStart(newDev)	

	######################
	def triggerStartProcessing(self, trigger):
		self.debugLog(u"<<-- entering triggerStartProcessing: %s (%d)" % (trigger.name, trigger.id))
		self.iPresence.triggerStart(trigger)
		self.debugLog(u"exiting triggerStartProcessing -->>")

	def triggerStopProcessing(self, trigger):
		self.debugLog(u"<<-- entering triggerStopProcessing: %s (%d)" % (trigger.name, trigger.id))
		self.iPresence.triggerStop(trigger)
		self.debugLog(u"exiting triggerStopProcessing -->>")

	########################################
		
	def runConcurrentThread(self):
		self.debugLog(u"runConcurrentThread called")
		self.iPresence.startComm()
		self.debugLog(u"Exiting Concurrent Thread")

	def stopConcurrentThread(self):
		self.debugLog(u"stopConcurrentThread called")
		self.iPresence.stopComm()
		self.debugLog(u"Exiting stopConcurrentThread")

	########################################
	# Prefs UI methods (works with PluginConfig.xml):
	######################
		
	# Validate the pluginConfig window after user hits OK
	# Returns False on failure, True on success
	#
	def validatePrefsConfigUi(self, valuesDict):	
		self.debugLog(u"validating Prefs called")
		
		errorMsgDict = indigo.Dict()
		err = False
		
		if valuesDict[u'minutesTillAway'].isdigit() == False or int(valuesDict[u'minutesTillAway']) < 2:	
			errorMsgDict[u'minutesTillAway'] = u"This must be 2 or higher."
			err = True
			
		if valuesDict[u'hoursTillVacation'].isdigit() == False:
			errorMsgDict[u'hoursTillVacation'] = u"This must be a valid number."
			err = True

		if valuesDict[u'configSyslogOrSNMP'] == u'snmp':
			if (len(valuesDict[u'snmpAddress']) == 0) or (valuesDict[u'snmpAddress'].count('.') < 1) :
				errorMsgDict[u'snmpAddress'] = u"Enter a valid IP or Bonjour address here."
				err = True
			
			if len(valuesDict[u'snmpCommunity']) == 0:
				errorMsgDict[u'snmpCommunity'] = u"Enter your Airport password or SNMP community string here."
				err = True
				
			if valuesDict[u'snmpAddress'].count(',') != valuesDict[u'snmpCommunity'].count(','):
				errorMsgDict[u'snmpCommunity'] = u"The number of passwords must be equal to the number of addresses."
				err = True
				
		if err == True:
			return (False, valuesDict, errorMsgDict)
			
		# Tell iPresence module to reread it's config
		self.iPresence.configRead = False
		# User choices look good, so return True (client will then close the dialog window).
		return (True, valuesDict)


	def validateDeviceConfigUi(self, valuesDict, typeId, devId):
		self.debugLog(u"validating Device Config called")
		#self.debugLog(u"Type: %s, Id: %s, Dict: %s" % (typeId, devId, valuesDict))
		
		if (u'configSyslogOrSNMP' in self.pluginPrefs) and (self.pluginPrefs[u'configSyslogOrSNMP'] == u'snmp'):
			#check if mac
			macAdr = valuesDict[u'phoneID']
			if macAdr.count(':') != 5:
				errorMsgDict = indigo.Dict()
				errorMsgDict[u'phoneID'] = u"This needs to be a valid MAC address."
				return (False, valuesDict, errorMsgDict)
			
		return (True, valuesDict)	
